/*
 * Copyright 2009-2012 the Fess Project and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package jp.sf.fess.action.admin;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.annotation.Resource;

import jp.sf.fess.Constants;
import jp.sf.fess.crud.util.SAStrutsUtil;
import jp.sf.fess.form.admin.SystemForm;
import jp.sf.fess.helper.SystemHelper;
import jp.sf.fess.helper.WebManagementHelper;
import jp.sf.fess.solr.SolrServerGroup;
import jp.sf.fess.solr.SolrServerManager;
import jp.sf.fess.task.CrawlTask;
import jp.sf.fess.util.FessProperties;

import org.apache.solr.client.solrj.SolrQuery;
import org.apache.solr.client.solrj.SolrQuery.ORDER;
import org.apache.solr.client.solrj.response.FacetField;
import org.apache.solr.client.solrj.response.QueryResponse;
import org.seasar.framework.container.SingletonS2Container;
import org.seasar.framework.util.StringUtil;
import org.seasar.struts.annotation.ActionForm;
import org.seasar.struts.annotation.Execute;
import org.seasar.struts.exception.ActionMessagesException;
import org.seasar.struts.taglib.S2Functions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class SystemAction implements Serializable {
    private static final Logger logger = LoggerFactory
            .getLogger(SystemAction.class);

    private static final long serialVersionUID = 1L;

    @ActionForm
    @Resource
    protected SystemForm systemForm;

    @Resource
    protected FessProperties solrServerProperties;

    @Resource
    protected SolrServerManager solrServerManager;

    @Resource
    protected WebManagementHelper webManagementHelper;

    @Resource
    protected SystemHelper systemHelper;

    public String getHelpLink() {
        return systemHelper.getHelpLink("system");
    }

    protected String showIndex(final boolean redirect) {
        final String[] serverNames = solrServerManager.getSolrServerNames();
        for (final String serverName : serverNames) {
            final HashMap<String, String> map = new HashMap<String, String>(2);
            map.put("serverName", serverName.replace('.', '/'));
            String status = solrServerProperties.getProperty(serverName
                    + ".status");
            if (StringUtil.isBlank(status)) {
                status = Constants.ACTIVE;
            }
            map.put("status", status);
            systemForm.serverStatusList.add(map);
        }

        systemForm.currentServerForSelect = solrServerProperties
                .getProperty(Constants.CURRENT_SERVER_FOR_SELECT);
        systemForm.currentServerStatusForSelect = solrServerProperties
                .getProperty(systemForm.currentServerForSelect + ".status");
        if (StringUtil.isBlank(systemForm.currentServerStatusForSelect)) {
            systemForm.currentServerStatusForSelect = Constants.ACTIVE;
        }

        systemForm.currentServerForUpdate = solrServerProperties
                .getProperty(Constants.CURRENT_SERVER_FOR_UPDATE);
        systemForm.currentServerStatusForUpdate = solrServerProperties
                .getProperty(systemForm.currentServerForUpdate + ".status");
        if (StringUtil.isBlank(systemForm.currentServerStatusForUpdate)) {
            systemForm.currentServerStatusForUpdate = Constants.ACTIVE;
        }

        if (redirect) {
            return "index?redirect=true";
        } else {
            return "index.jsp";
        }
    }

    @Execute(validator = false)
    public String index() {
        return showIndex(false);
    }

    @Execute(validator = true, input = "index.jsp")
    public String update() {
        // server status
        for (final Map<String, String> statusMap : systemForm.serverStatusList) {
            for (final Map.Entry<String, String> entry : statusMap.entrySet()) {
                final String value = entry.getValue();
                solrServerProperties.setProperty(
                        entry.getKey().replace('/', '.') + ".status", value);
                if (Constants.ACTIVE.equals(value)) {
                    final String groupKey = entry.getKey().split("/")[0]
                            + ".status";
                    if (!Constants.ACTIVE.equals(solrServerProperties
                            .getProperty(groupKey))) {
                        solrServerProperties.remove(groupKey);
                    }
                }
            }
        }

        try {
            solrServerProperties.store();
            SAStrutsUtil.addSessionMessage("success.update_solr_params");
        } catch (final Exception e) {
            logger.error("Failed to update solr parameters.", e);
            throw new ActionMessagesException(
                    "errors.failed_to_update_solr_params", e);
        }

        return showIndex(true);

    }

    @Execute(validator = true, input = "index")
    public String commit() {
        if (systemHelper.isCrawlProcessRunning()) {
            throw new ActionMessagesException(
                    "errors.failed_to_start_solr_process_because_of_running");
        }
        final SolrServerGroup serverGroup = solrServerManager
                .getSolrServerGroup(systemForm.groupName);
        if (serverGroup == null) {
            throw new ActionMessagesException(
                    "errors.failed_to_commit_solr_index");
        } else {
            final Thread thread = new Thread(new Runnable() {
                @Override
                public void run() {
                    if (systemHelper.readyCrawlProcess()) {
                        final long execTime = System.currentTimeMillis();
                        try {
                            serverGroup.commit();
                            if (logger.isInfoEnabled()) {
                                logger.info("[EXEC TIME] index commit time: "
                                        + (System.currentTimeMillis() - execTime)
                                        + "ms");
                            }
                        } catch (final Exception e) {
                            logger.error("Failed to commit index.", e);
                        } finally {
                            systemHelper.finishCrawlProcess();
                        }
                    } else {
                        if (logger.isInfoEnabled()) {
                            logger.info("could not start index cleanup process"
                                    + " because of running solr process.");
                        }
                    }
                }
            });
            thread.start();
            SAStrutsUtil.addSessionMessage("success.commit_solr_index");
        }
        return showIndex(true);
    }

    @Execute(validator = true, input = "index")
    public String optimize() {
        if (systemHelper.isCrawlProcessRunning()) {
            throw new ActionMessagesException(
                    "errors.failed_to_start_solr_process_because_of_running");
        }
        final SolrServerGroup serverGroup = solrServerManager
                .getSolrServerGroup(systemForm.groupName);
        if (serverGroup == null) {
            throw new ActionMessagesException(
                    "errors.failed_to_optimize_solr_index");
        } else {
            final Thread thread = new Thread(new Runnable() {
                @Override
                public void run() {
                    if (systemHelper.readyCrawlProcess()) {
                        final long execTime = System.currentTimeMillis();
                        try {
                            serverGroup.optimize();
                            if (logger.isInfoEnabled()) {
                                logger.info("[EXEC TIME] index optimize time: "
                                        + (System.currentTimeMillis() - execTime)
                                        + "ms");
                            }
                        } catch (final Exception e) {
                            logger.error("Failed to optimize index.", e);
                        } finally {
                            systemHelper.finishCrawlProcess();
                        }
                    } else {
                        if (logger.isInfoEnabled()) {
                            logger.info("could not start index cleanup process"
                                    + " because of running solr process.");
                        }
                    }
                }
            });
            thread.start();
            SAStrutsUtil.addSessionMessage("success.optimize_solr_index");
        }
        return showIndex(true);
    }

    @Execute(validator = true, input = "index")
    public String delete() {
        String deleteQuery;
        if ("*".equals(systemForm.sessionId)) {
            deleteQuery = "*:*";
        } else {
            deleteQuery = "segment:" + systemForm.sessionId;
        }
        return deleteByQuery(deleteQuery);
    }

    @Execute(validator = true, input = "index")
    public String confirmByUrl() {
        final String confirmQuery = "url:\"" + systemForm.deleteUrl + "\"";
        return "/admin/searchList/search?query=" + S2Functions.u(confirmQuery)
                + "&redirect=true";
    }

    @Execute(validator = true, input = "index")
    public String deleteByUrl() {
        final String deleteUrl = systemForm.deleteUrl;
        final String deleteQuery = "url:\"" + deleteUrl + "\"";
        return deleteByQuery(deleteQuery);
    }

    private String deleteByQuery(final String deleteQuery) {
        if (systemHelper.isCrawlProcessRunning()) {
            throw new ActionMessagesException(
                    "errors.failed_to_start_solr_process_because_of_running");
        }
        final SolrServerGroup serverGroup = solrServerManager
                .getSolrServerGroup(systemForm.groupName);
        if (serverGroup == null) {
            throw new ActionMessagesException(
                    "errors.failed_to_delete_solr_index");
        } else {
            final Thread thread = new Thread(new Runnable() {
                @Override
                public void run() {
                    if (systemHelper.readyCrawlProcess()) {
                        final long execTime = System.currentTimeMillis();
                        try {
                            serverGroup.deleteByQuery(deleteQuery);
                            serverGroup.optimize();
                            if (logger.isInfoEnabled()) {
                                logger.info("[EXEC TIME] index cleanup time: "
                                        + (System.currentTimeMillis() - execTime)
                                        + "ms");
                            }
                        } catch (final Exception e) {
                            logger.error("Failed to delete index (query="
                                    + deleteQuery + ").", e);
                        } finally {
                            systemHelper.finishCrawlProcess();
                        }
                    } else {
                        if (logger.isInfoEnabled()) {
                            logger.info("could not start index cleanup process"
                                    + " because of running solr process.");
                        }
                    }
                }
            });
            thread.start();
            SAStrutsUtil.addSessionMessage("success.delete_solr_index");
        }
        return showIndex(true);
    }

    @Execute(validator = true, input = "index")
    public String start() {
        final SystemHelper systemHelper = SingletonS2Container
                .getComponent("systemHelper");
        if (!systemHelper.isCrawlProcessRunning()) {
            final CrawlTask crawlTask = SingletonS2Container
                    .getComponent(CrawlTask.class);
            new Thread(new Runnable() {
                @Override
                public void run() {
                    crawlTask.doExecute();
                }
            }).start();
            SAStrutsUtil.addSessionMessage("success.start_crawl_process");
        } else {
            SAStrutsUtil
                    .addSessionMessage("success.failed_to_start_crawl_process");
        }
        return showIndex(true);
    }

    @Execute(validator = true, input = "index")
    public String stop() {
        final SystemHelper systemHelper = SingletonS2Container
                .getComponent("systemHelper");
        if (systemHelper.isCrawlProcessRunning()) {
            systemHelper.setForceStop(true);
            SAStrutsUtil.addSessionMessage("success.stopping_crawl_process");
        } else {
            SAStrutsUtil.addSessionMessage("errors.no_running_crawl_process");
        }
        return showIndex(true);
    }

    @Execute(validator = true, input = "index.jsp")
    public String startSolrInstance() {
        try {
            webManagementHelper.start(systemForm.solrInstanceName);
            SAStrutsUtil.addSessionMessage("success.starting_solr_instance");
        } catch (final Exception e) {
            logger.error("Failed to start a solr instance: "
                    + systemForm.solrInstanceName, e);
            throw new ActionMessagesException(
                    "errors.failed_to_start_solr_instance");
        }
        return showIndex(true);
    }

    @Execute(validator = true, input = "index.jsp")
    public String stopSolrInstance() {
        try {
            webManagementHelper.stop(systemForm.solrInstanceName);
            SAStrutsUtil.addSessionMessage("success.stopping_solr_instance");
        } catch (final Exception e) {
            logger.error("Failed to stop a solr instance: "
                    + systemForm.solrInstanceName, e);
            throw new ActionMessagesException(
                    "errors.failed_to_stop_solr_instance");
        }
        return showIndex(true);
    }

    @Execute(validator = true, input = "index.jsp")
    public String reloadSolrInstance() {
        try {
            webManagementHelper.reload(systemForm.solrInstanceName);
            SAStrutsUtil.addSessionMessage("success.reloading_solr_instance");
        } catch (final Exception e) {
            logger.error("Failed to reload a solr instance: "
                    + systemForm.solrInstanceName, e);
            throw new ActionMessagesException(
                    "errors.failed_to_reload_solr_instance");
        }
        return showIndex(true);
    }

    public List<Map<String, Object>> getGroupActionItems() {
        final List<Map<String, Object>> groupActionItems = new ArrayList<Map<String, Object>>();
        try {
            for (final String groupName : solrServerManager
                    .getSolrServerGroupNames()) {
                final Map<String, Object> map = new HashMap<String, Object>();
                map.put("groupName", groupName);
                final SessionIdList<Map<String, String>> sessionIdList = getSessionIdList(groupName);
                map.put("sessionIdItems", sessionIdList);
                map.put("totalCount", sessionIdList.getTotalCount());
                groupActionItems.add(map);
            }
        } catch (final Exception e) {
            logger.info("could not get server groups.", e);
        }
        return groupActionItems;
    }

    protected SessionIdList<Map<String, String>> getSessionIdList(
            final String groupName) {
        final SessionIdList<Map<String, String>> sessionIdList = new SessionIdList<Map<String, String>>();

        SolrServerGroup serverGroup;
        try {
            serverGroup = solrServerManager.getSolrServerGroup(groupName);
        } catch (final Exception e) {
            if (logger.isInfoEnabled()) {
                logger.info(e.getMessage());
            }
            return sessionIdList;
        }

        final SolrQuery query = new SolrQuery();
        query.setQuery("*:*");
        query.setFacet(true);
        query.addFacetField("segment");
        query.addSortField("segment", ORDER.desc);

        final QueryResponse queryResponse = serverGroup.query(query);
        final List<FacetField> facets = queryResponse.getFacetFields();
        for (final FacetField facet : facets) {
            final List<FacetField.Count> facetEntries = facet.getValues();
            if (facetEntries != null) {
                for (final FacetField.Count fcount : facetEntries) {
                    final Map<String, String> map = new HashMap<String, String>(
                            3);
                    map.put("label",
                            fcount.getName() + " (" + fcount.getCount() + ")");
                    map.put("value", fcount.getName());
                    map.put("count", Long.toString(fcount.getCount()));
                    sessionIdList.add(map);
                    sessionIdList.addTotalCount(fcount.getCount());
                }
            }
        }
        return sessionIdList;
    }

    public List<Map<String, String>> getSolrInstanceList() {
        final List<String> solrInstanceNameList = webManagementHelper
                .getSolrInstanceNameList();
        final List<Map<String, String>> solrInstanceList = new ArrayList<Map<String, String>>();
        for (final String solrInstanceName : solrInstanceNameList) {
            final Map<String, String> map = new HashMap<String, String>();
            map.put("name", solrInstanceName);
            map.put("status", webManagementHelper.getStatus(solrInstanceName));
            solrInstanceList.add(map);
        }
        return solrInstanceList;
    }

    public boolean isSolrProcessRunning() {
        return systemHelper.isCrawlProcessRunning();
    }

    public String getRunningSessionId() {
        return systemHelper.getSessionId();
    }

    private static class SessionIdList<E> extends ArrayList<E> {

        private static final long serialVersionUID = 1L;

        private long totalCount = 0;

        public void addTotalCount(final long count) {
            totalCount += count;
        }

        public long getTotalCount() {
            return totalCount;
        }

    }
}